@@ -1,6 +1,6 @@
 import re
 
-from .constants import SHA_COINS, SCRYPT_COINS, COIN_SYMBOL_SET, COIN_SYMBOL_MAPPINGS, FIRST4_MKEY_CS_MAPPINGS_UPPER, UNIT_CHOICES, UNIT_MAPPINGS
+from .constants import SHA_COINS, SCRYPT_COINS, ETHASH_COINS, COIN_SYMBOL_SET, COIN_SYMBOL_MAPPINGS, FIRST4_MKEY_CS_MAPPINGS_UPPER, UNIT_CHOICES, UNIT_MAPPINGS
 from .crypto import script_to_address
 
 from bitcoin import safe_from_hex, deserialize
@@ -18,39 +18,48 @@ def format_output(num, output_type):
         return '{0:,.5f}'.format(num)
     elif output_type == 'bit':
         return '{0:,.2f}'.format(num)
-    elif output_type == 'satoshi':
+    elif output_type in ['satoshi', 'wei']:
         return '{:,}'.format(int(num))
+    elif output_type == 'ether':
+        return '{0:,.18f}'.format(num)
     else:
         raise Exception('Invalid Unit Choice: %s' % output_type)
 
 
-def to_satoshis(input_quantity, input_type):
-    ''' convert to satoshis, no rounding '''
+def to_base_unit(input_quantity, input_type):
+    ''' convert to satoshis or wei, no rounding '''
     assert input_type in UNIT_CHOICES, input_type
 
     # convert to satoshis
     if input_type in ('btc', 'mbtc', 'bit'):
-        satoshis = float(input_quantity) * float(UNIT_MAPPINGS[input_type]['satoshis_per'])
-    elif input_type == 'satoshi':
-        satoshis = input_quantity
+        base_unit = float(input_quantity) * float(UNIT_MAPPINGS[input_type]['satoshis_per'])
+    elif input_type in ('ether', 'gwei'):
+        base_unit = float(input_quantity) * float(UNIT_MAPPINGS[input_type]['wei_per'])
+    elif input_type in ['satoshi', 'wei']:
+        base_unit = input_quantity
     else:
         raise Exception('Invalid Unit Choice: %s' % input_type)
 
-    return int(satoshis)
+    return int(base_unit)
 
 
-def from_satoshis(input_satoshis, output_type):
+def from_base_unit(input_base, output_type):
     # convert to output_type,
     if output_type in ('btc', 'mbtc', 'bit'):
-        return input_satoshis / float(UNIT_MAPPINGS[output_type]['satoshis_per'])
-    elif output_type == 'satoshi':
-        return int(input_satoshis)
+        return input_base / float(UNIT_MAPPINGS[output_type]['satoshis_per'])
+    elif output_type in ('ether', 'gwei'):
+        return input_base / float(UNIT_MAPPINGS[output_type]['wei_per'])
+    elif output_type in ['satoshi', 'wei']:
+        return int(input_base)
     else:
         raise Exception('Invalid Unit Choice: %s' % output_type)
 
 
 def satoshis_to_btc(satoshis):
-    return from_satoshis(input_satoshis=satoshis, output_type='btc')
+    return from_base_unit(input_base=satoshis, output_type='btc')
+
+def wei_to_ether(wei):
+    return from_base_unit(input_base=wei, output_type='ether')
 
 
 def get_curr_symbol(coin_symbol, output_type):
@@ -104,15 +113,15 @@ def format_crypto_units(input_quantity, input_type, output_type, coin_symbol=Non
     assert output_type in UNIT_CHOICES, output_type
     if print_cs:
         assert is_valid_coin_symbol(coin_symbol=coin_symbol), coin_symbol
-    assert type(round_digits) is int
+    assert isinstance(round_digits, int)
 
-    satoshis_float = to_satoshis(input_quantity=input_quantity, input_type=input_type)
+    base_unit_float = to_base_unit(input_quantity=input_quantity, input_type=input_type)
 
     if round_digits:
-        satoshis_float = round(satoshis_float, -1*round_digits)
+        base_unit_float = round(base_unit_float, -1*round_digits)
 
-    output_quantity = from_satoshis(
-            input_satoshis=satoshis_float,
+    output_quantity = from_base_unit(
+            input_base=base_unit_float,
             output_type=output_type,
             )
 
@@ -124,7 +133,7 @@ def format_crypto_units(input_quantity, input_type, output_type, coin_symbol=Non
         # add thousands separator and appropriate # of decimals
         output_quantity_formatted = format_output(num=output_quantity, output_type=output_type)
 
-    if safe_trimming and output_type not in ('satoshi', 'bit'):
+    if safe_trimming and output_type not in ('satoshi', 'wei', 'bit'):
         output_quantity_formatted = safe_trim(qty_as_string=output_quantity_formatted)
 
     if print_cs:
@@ -165,35 +174,38 @@ def get_txn_outputs(raw_tx_hex, output_addr_list, coin_symbol):
     # Defensive checks:
     err_msg = 'Library not able to parse %s transactions' % coin_symbol
     assert lib_can_deserialize_cs(coin_symbol), err_msg
-    assert type(output_addr_list) in (list, tuple)
+    assert isinstance(output_addr_list, (list, tuple))
     for output_addr in output_addr_list:
-        assert is_valid_address(output_addr), output_addr
+        assert is_valid_address_for_coinsymbol(output_addr, coin_symbol), output_addr
 
     output_addr_set = set(output_addr_list)  # speed optimization
 
     outputs = []
     deserialized_tx = deserialize(str(raw_tx_hex))
     for out in deserialized_tx.get('outs', []):
+        output = {'value': out['value']}
 
-        # determine if the address is a pubkey address or a script address
+        # determine if the address is a pubkey address, script address, or op_return
         pubkey_addr = script_to_address(out['script'],
-                vbyte=COIN_SYMBOL_MAPPINGS[coin_symbol]['vbyte_pubkey'])
+            vbyte=COIN_SYMBOL_MAPPINGS[coin_symbol]['vbyte_pubkey'])
         script_addr = script_to_address(out['script'],
-                vbyte=COIN_SYMBOL_MAPPINGS[coin_symbol]['vbyte_script'])
+            vbyte=COIN_SYMBOL_MAPPINGS[coin_symbol]['vbyte_script'])
+        nulldata = out['script'] if out['script'][0:2] == '6a' else None
         if pubkey_addr in output_addr_set:
             address = pubkey_addr
+            output['address'] = address
         elif script_addr in output_addr_set:
             address = script_addr
+            output['address'] = address
+        elif nulldata:
+            output['script'] = nulldata
+            output['script_type'] = 'null-data'
         else:
             raise Exception('Script %s Does Not Contain a Valid Output Address: %s' % (
                 out['script'],
                 output_addr_set,
                 ))
 
-        output = {
-                'value': out['value'],
-                'address': address,
-                }
         outputs.append(output)
     return outputs
 
@@ -207,7 +219,8 @@ def compress_txn_outputs(txn_outputs):
         {'1abc...': 12345, '1def': 54321, ...}
     '''
     result_dict = {}
-    for txn_output in txn_outputs:
+    outputs = (output for output in txn_outputs if output.get('address'))
+    for txn_output in outputs:
         if txn_output['address'] in result_dict:
             result_dict[txn_output['address']] += txn_output['value']
         else:
@@ -228,7 +241,7 @@ def get_txn_outputs_dict(raw_tx_hex, output_addr_list, coin_symbol):
 def compress_txn_inputs(txn_inputs):
     result_dict = {}
     for txn_input in txn_inputs:
-        if 'addresses' in txn_input:
+        if 'addresses' in txn_input and 'output_value' in txn_input:
             # coinbase tx has no address
             address = txn_input['addresses'][0]
             if address in result_dict:
@@ -244,10 +257,11 @@ def estimate_satoshis_transacted(inputs, outputs):
     # total sent less fees and "known" change
     satoshis_transacted = 0
     for output in outputs:
-        if len(output['addresses']) == 0:
+        output_addresses = output.get('addresses')
+        if not output_addresses:
             # null data - count it for the estimate
             satoshis_transacted += output['value']
-        elif output['addresses'][0] not in inputs_compressed:
+        elif output_addresses and output_addresses[0] not in inputs_compressed:
             # not change
             satoshis_transacted += output['value']
 
@@ -280,7 +294,7 @@ def get_blockcypher_walletname_from_mpub(mpub, subchain_indices=[]):
 
     if subchain_indices:
         mpub += ','.join([str(x) for x in subchain_indices]).encode('utf-8')
-    return sha256(mpub).hexdigest()[:25]
+    return 'X%s' % sha256(mpub).hexdigest()[:24]
 
 
 def is_valid_wallet_name(wallet_name):
@@ -355,7 +369,7 @@ def is_valid_block_num(block_num):
         return False
 
     # hackey approximation
-    return 0 <= bn_as_int <= 10**8
+    return 0 <= bn_as_int <= 10**9
 
 
 def is_valid_sha_block_hash(block_hash):
@@ -366,6 +380,10 @@ def is_valid_scrypt_block_hash(block_hash):
     " Unfortunately this is indistiguishable from a regular hash "
     return is_valid_hash(block_hash)
 
+def is_valid_ethash_block_hash(block_hash):
+    " Unfortunately this is indistiguishable from a regular hash "
+    return is_valid_hash(block_hash)
+
 
 def is_valid_sha_block_representation(block_representation):
     return is_valid_block_num(block_representation) or is_valid_sha_block_hash(block_representation)
@@ -374,6 +392,8 @@ def is_valid_sha_block_representation(block_representation):
 def is_valid_scrypt_block_representation(block_representation):
     return is_valid_block_num(block_representation) or is_valid_scrypt_block_hash(block_representation)
 
+def is_valid_ethash_block_representation(block_representation):
+    return is_valid_block_num(block_representation) or is_valid_ethash_block_hash(block_representation)
 
 def is_valid_bcy_block_representation(block_representation):
     block_representation = str(block_representation)
@@ -398,6 +418,10 @@ def is_valid_block_representation(block_representation, coin_symbol):
             return is_valid_sha_block_representation(block_representation)
     elif coin_symbol in SCRYPT_COINS:
         return is_valid_scrypt_block_representation(block_representation)
+    elif coin_symbol in ETHASH_COINS:
+        return is_valid_ethash_block_representation(block_representation)
+    else:
+        return True
 
 
 # Coin Symbol #
@@ -452,12 +476,24 @@ def crypto_address_valid(bc):
 
 
 def is_valid_address(b58_address):
+    # TODO deeper validation of a bech32 address
+    if b58_address.startswith('bc1') or b58_address.startswith('ltc1') or b58_address.startswith('tltc1') or b58_address.startswith('tb1'):
+        return True
+
     try:
         return crypto_address_valid(b58_address)
     except:
         # handle edge cases like an address too long to decode
         return False
 
+def is_valid_eth_address(addr):
+    if addr.startswith('0x'):
+        addr = addr[2:].strip()
+
+    if len(addr) != 40:
+        return False
+
+    return uses_only_hash_chars(addr)
 
 def is_valid_address_for_coinsymbol(b58_address, coin_symbol):
     '''
@@ -466,6 +502,13 @@ def is_valid_address_for_coinsymbol(b58_address, coin_symbol):
     '''
     assert is_valid_coin_symbol(coin_symbol)
 
+    # TODO deeper validation of a bech32 address
+    if b58_address.startswith(COIN_SYMBOL_MAPPINGS[coin_symbol]['bech32_prefix']):
+        return True
+
+    if coin_symbol == 'eth':
+        return is_valid_eth_address(b58_address)
+
     if b58_address[0] in COIN_SYMBOL_MAPPINGS[coin_symbol]['address_first_char_list']:
         if is_valid_address(b58_address):
             return True
